// @ts-check

class Vector2D {

    /**
     * 
     * @param {number} x 
     * @param {number} y 
     */
    constructor(x, y) {
        this.x = x;
        this.y = y;
    }

    /**
     * @returns {Vector2D}
     */
    negative() {
        return new Vector2D(-this.x, -this.y);
    }

    /**
     * @param {Vector2D|number} v
     * @returns {Vector2D}
     */
    add(v) {
        return v instanceof Vector2D ?
            new Vector2D(this.x + v.x, this.y + v.y)
            : new Vector2D(this.x + v, this.y + v);
    }

    /**
     * @param {Vector2D|number} v
     * @returns {Vector2D}
     */
    substract(v) {
        return v instanceof Vector2D ?
            new Vector2D(this.x - v.x, this.y - v.y)
            : new Vector2D(this.x - v, this.y - v);
    }

    /**
     * @param {Vector2D|number} v
     * @returns {Vector2D}
     */
    multiply(v) {
        return v instanceof Vector2D ?
            new Vector2D(this.x * v.x, this.y * v.y)
            : new Vector2D(this.x * v, this.y * v);
    }

    /**
     * @param {Vector2D|number} v
     * @returns {Vector2D}
     */
    divide(v) {
        return v instanceof Vector2D ?
            new Vector2D(this.x / v.x, this.y / v.y)
            : new Vector2D(this.x / v, this.y / v);
    }

    /**
     * 
     * @param {Vector2D} v
     * @returns {boolean} 
     */
    equals(v) {
        return v instanceof Vector2D ?
            this.x === v.x && this.y === v.y
            : false
    }

    /**
     * @param {Vector2D} v
     * @returns {number}
     */
    dot(v) {
        return this.x * v.x + this.y * v.y;
    }

    /**
     * @returns {number}
     */
    length() {
        return Math.sqrt(this.dot(this));
    }

    /**
     * @param {Vector2D} v
     * @returns {number}
     */
    angleTo(v) {
        return Math.acos(this.dot(v) / (this.length() * v.length()));
    }

    /**
     * @returns {Vector2D}
     */
    unit(){
        return this.divide( this.length() )
    }

    /*** Copie methods ***/

    /**
     * @returns {Vector2D}
     */
    clone() {
        return new Vector2D(this.x, this.y);
    }

    /**
     * @returns {Array}
     */
    toArray() {
        return [this.x, this.y];
    }

    /**
     * @returns {Object}
     */
    toJSON() {
        return { x: this.x, y: this.y };
    }

    /* STATIC METHODS */

    /**
     * @param {Object} json
     * @returns {Vector2D}
     */
    static fromJSON(json) {
        return new Vector2D(json.x, json.y);
    }

    /**
     * @param {Array} array
     * @returns {Vector2D}
     */
    static fromArray(array) {
        return new Vector2D(array[0], array[1]);
    }
}

module.exports = Vector2D;
